/*
  sdlhack - Force SDL games to minimize when Ctrl-Z is pressed

  Copyright (C) 2010 Jared Stafford (jspenguin@gmail.com)

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdlib.h>
#include <stdio.h>
#include <dlfcn.h>
#include <SDL.h>

#define VERSION "1.4"

/* dlopen() handle for libSDL */
static void* sdlhandle;

/* Function pointers to real SDL library. */
static int (*orig_SDL_Init)(Uint32 mask);
static int (*orig_SDL_WasInit)(Uint32 mask);
static void (*orig_SDL_SetError)(const char* fmt, ...);
static SDL_Surface* (*orig_SDL_SetVideoMode)(int width, int height, int bpp, Uint32 flags);
static void (*orig_SDL_SetEventFilter)(SDL_EventFilter filter);
static SDL_EventFilter (*orig_SDL_GetEventFilter)();
static char* (*orig_SDL_GetKeyName)(SDLKey key);
static void (*orig_SDL_WM_IconifyWindow)(); 
static SDL_GrabMode (*orig_SDL_WM_GrabInput)(SDL_GrabMode mode);

/* These are only defined if joystick support is not disabled */
static int (*orig_SDL_NumJoysticks)(void);
static SDL_Joystick *(*orig_SDL_JoystickOpen)(int device_index);
static const char *(*orig_SDL_JoystickName)(int device_index);
static int (*orig_SDL_JoystickOpened)(int device_index);
static int (*orig_SDL_JoystickEventState)(int state);
static int (*orig_SDL_OpenAudio)(SDL_AudioSpec* desired, SDL_AudioSpec* obtained);
static int (*orig_SDL_CloseAudio)(void);

/* Event filter set by the application, if any. */
static SDL_EventFilter app_event_filter;

/* Will be SDL_GRAB_ON if the app wants input grabbed */
static SDL_GrabMode app_grabmode = SDL_GRAB_OFF;

static FILE* audio_tee_file;
static void (SDLCALL *app_audio_callback)(void *userdata, Uint8 *stream, int len);

/* Will be 1 if we forced app to minimize */
static int force_grab_off = 0;

/* */
static SDLMod trigger_mods = 0;
static SDLKey trigger_key = 0;

static char wav_header1[] = { 0x52, 0x49, 0x46, 0x46, 0,0,0,0, 0x57, 0x41, 0x56, 0x45,
		  0x66, 0x6d, 0x74, 0x20, 0x10, 0x00, 0x00, 0x00 };
static char wav_header2[] = { 0x64, 0x61, 0x74, 0x61, 0,0,0,0 };

static int force_res_x = -1;
static int force_res_y = -1;

static SDLCALL int our_event_filter(const SDL_Event* evt);

static void* load_sym(const char* name) {
    void* sym = dlsym(sdlhandle, name);
    if (!sym) {
        printf("failed loading %s, err = %s\n", name, dlerror());
        exit(1);
    }
    printf ("  %-22s = %p\n", name, sym);
    return sym;
}

static void sdlhack_init() {
    char* joydisable;
    char* forceres;

    sdlhandle = dlopen("libSDL-1.2.so.0", RTLD_NOW);
    if (!sdlhandle) {
        fprintf(stderr, "Could not load SDL: %s\n", dlerror());
        /* What to do? The game called SDL_Init, but SDL can't load?
           Probably best to just bail. */
        exit(1);
    }
    printf ("SDL iconify hack version " VERSION " initializing, dlopen handle = %p\n", sdlhandle);

#define SYM_INIT(sym) orig_##sym = load_sym(#sym);

    SYM_INIT(SDL_Init);
    SYM_INIT(SDL_WasInit);
    SYM_INIT(SDL_SetError);

    SYM_INIT(SDL_SetVideoMode);
    SYM_INIT(SDL_SetEventFilter);
    SYM_INIT(SDL_GetEventFilter);
    SYM_INIT(SDL_WM_IconifyWindow);
    SYM_INIT(SDL_WM_GrabInput);
    SYM_INIT(SDL_GetKeyName);
    SYM_INIT(SDL_OpenAudio);
    SYM_INIT(SDL_CloseAudio);

    if ((joydisable = getenv("SDLHACK_DISABLE_JOYSTICK")) == NULL ||
        atoi(joydisable) == 0) {
        
        SYM_INIT(SDL_NumJoysticks);
        SYM_INIT(SDL_JoystickOpen);
        SYM_INIT(SDL_JoystickName);
        SYM_INIT(SDL_JoystickOpened);
        SYM_INIT(SDL_JoystickEventState);
    } else {
        printf("joystick disabled\n");
    }
    if ((forceres = getenv("SDLHACK_FORCE_RES")) != NULL) {
        if (sscanf(forceres, "%dx%d", &force_res_x, &force_res_y) < 2) {
            force_res_x = force_res_y = -1;
        } else {
            printf("forcing resolution to %dx%d\n", force_res_x, force_res_y);
        }
    }
}

/* Must be called after SDL_init */

static void init_trigger() {
    char* trigger_env;
    if ((trigger_env = getenv("SDLHACK_TRIGGER_KEY")) != NULL) {
        char* cp = trigger_env;
        char* end;
        SDLKey curkey;

        trigger_mods = 0;
        trigger_key = 0;

        while ((end = strchr(cp, '-')) != NULL) {
#define MOD_IS(text) ((end - cp) == (sizeof(#text) - 1) && !strncasecmp(cp, #text, (sizeof(#text) - 1)))
            /* special case: ctrl-- means ctrl + '-' */
            if (end == cp && (end[1] == '\0'))
                break;
            if (MOD_IS(ctrl)) {
                trigger_mods |= KMOD_LCTRL;
            } else if (MOD_IS(shift)) {
                trigger_mods |= KMOD_LSHIFT;
            } else if (MOD_IS(alt)) {
                trigger_mods |= KMOD_LALT;
            } else if (MOD_IS(meta)) {
                trigger_mods |= KMOD_LMETA;
            } else {
                printf("Warning: unknown modifier: ");
                fwrite(cp, (end - cp), 1, stdout);
                putchar('\n');
            }
            cp = end + 1;
        }
        if (sscanf(cp, "key_%u", &trigger_key) != 1) {
            /* Rest of the string now contains key name */
            for (curkey = 1; curkey < SDLK_LAST; curkey++) {
                const char* keyname = orig_SDL_GetKeyName(curkey);
                if (!strcasecmp(keyname, cp)) {
                    trigger_key = curkey;
                    break;
                }
            }
        }
        if (!trigger_key) {
            printf("Warning: unknown key \"%s\", using default\n", cp);
            trigger_key = SDLK_z;
        }
    } else {
        trigger_key = SDLK_z;
        trigger_mods = KMOD_LCTRL;
    }
    

    printf("Trigger key is %s%s%s%s%s\n", 
           (trigger_mods & KMOD_LCTRL ?  "ctrl-"  : ""),
           (trigger_mods & KMOD_LALT ?   "alt-"   : ""),
           (trigger_mods & KMOD_LSHIFT ? "shift-" : ""),
           (trigger_mods & KMOD_LMETA ?  "meta-"  : ""),
           orig_SDL_GetKeyName(trigger_key));

#undef SYM_INIT

}

static void check_event_filter() {
    if (orig_SDL_WasInit(SDL_INIT_VIDEO) && orig_SDL_GetEventFilter() != our_event_filter)
        orig_SDL_SetEventFilter(our_event_filter);
}

int SDL_Init(Uint32 mask) {
    int rc;
    if (!sdlhandle)
        sdlhack_init();
 
    if (!orig_SDL_NumJoysticks)
        mask &= ~SDL_INIT_JOYSTICK;

    rc = orig_SDL_Init(mask);

    if (rc == 0)
        check_event_filter();

    if (orig_SDL_WasInit(SDL_INIT_VIDEO) && !trigger_key)
        init_trigger();

    return rc;
}

static SDLCALL int our_event_filter(const SDL_Event* evt) {
    /* Check for Ctrl-Z */
    if (evt->type == SDL_KEYDOWN && 
        evt->key.keysym.sym == trigger_key) {

        /* We need to treat left & right mods equally */
        SDLMod testmods = 0;
        SDLMod mods = evt->key.keysym.mod;

        if (mods & KMOD_CTRL ) testmods |= KMOD_LCTRL;
        if (mods & KMOD_ALT  ) testmods |= KMOD_LALT;
        if (mods & KMOD_SHIFT) testmods |= KMOD_LSHIFT;
        if (mods & KMOD_META ) testmods |= KMOD_LMETA;


        if (testmods == trigger_mods) {
            printf("Iconify hack: Iconifying window\n");
            force_grab_off = 1;
            orig_SDL_WM_GrabInput(SDL_GRAB_OFF);

            orig_SDL_WM_IconifyWindow();

            /* eat the event */
            return 0;
        }
    }

    if (evt->type == SDL_ACTIVEEVENT && evt->active.gain) {
        if (force_grab_off) {
            force_grab_off = 0;
            printf("Iconify hack: Window restored\n");
            orig_SDL_WM_GrabInput(app_grabmode);
        }
    } 

    /* call application event filter */
    if (app_event_filter) return (*app_event_filter)(evt);
    return 1;
}


static void flushaudio();

static void SDLCALL tee_audio_callback(void *userdata, Uint8 *stream, int len) {
    app_audio_callback(userdata, stream, len);
    fwrite(stream, len, 1, audio_tee_file);
    flushaudio();
}

static void close_audio_tee(void) {
    if (audio_tee_file) {
        fclose(audio_tee_file);
        audio_tee_file = NULL;
    }
}

static void putshort(Uint8** p, unsigned short v) {
    *(*p)++ = v & 255;
    *(*p)++ = (v >> 8) & 255;
}

static void putint(Uint8** p, unsigned int v) {
    *(*p)++ = v & 255;
    *(*p)++ = (v >> 8) & 255;
    *(*p)++ = (v >> 16) & 255;
    *(*p)++ = (v >> 24) & 255;
}

static void flushaudio() {
    Uint8 *p, tbuf[4];
 
    long pos = ftell(audio_tee_file);
   
    fseek(audio_tee_file, 4, SEEK_SET);
    p = tbuf;
    putint(&p, pos - 8);
    fwrite(tbuf, 4, 1, audio_tee_file);
    
    fseek(audio_tee_file, 40, SEEK_SET);
    p = tbuf;
    putint(&p, pos - 44);
    fwrite(tbuf, 4, 1, audio_tee_file);

    fseek(audio_tee_file, pos, SEEK_SET);
    fflush(audio_tee_file);
}

int SDL_OpenAudio(SDL_AudioSpec* desired, SDL_AudioSpec* obtained) {
    int rc;
    char* fn;

    if (!sdlhandle)
        sdlhack_init();

    close_audio_tee();
    if ((fn = getenv("SDLHACK_TEE_AUDIO")) != NULL) {
        audio_tee_file = fopen(fn, "wb");
        if (audio_tee_file) {
            app_audio_callback = desired->callback;
            desired->callback = tee_audio_callback;
        } else{
            printf("Could not open audio output file for writing.\n");
        }
    }

    rc = orig_SDL_OpenAudio(desired, obtained);
    
    if (rc != 0) {
        close_audio_tee();
        return rc;
    }

    if (audio_tee_file) {
            Uint8 fmt_header[16];
            Uint8* p = fmt_header;
            int bps;

            fwrite(wav_header1, sizeof(wav_header1), 1, audio_tee_file);

            switch(obtained->format) {
                case AUDIO_U8:
                case AUDIO_S8:
                    bps = 1;
                default:
                    bps = 2;
            }
        
            putshort(&p, 1);
            putshort(&p, obtained->channels);
            putint(&p, obtained->freq);
            putint(&p, obtained->freq * obtained->channels * bps);
            putshort(&p, obtained->channels * bps);
            putshort(&p, bps * 8);
            fwrite(fmt_header, 16, 1, audio_tee_file);
            fwrite(wav_header2, sizeof(wav_header2), 1, audio_tee_file);
            printf("Opened %s, %dHz, %dch, %dbps\n", fn, obtained->freq, obtained->channels, bps);
    }
    return rc;
}

            
            
void SDL_CloseAudio(void) {
    close_audio_tee();
    orig_SDL_CloseAudio();
}



/* Recheck event filter after SDL_SetVideoMode*/
SDL_Surface* SDL_SetVideoMode(int width, int height, int bpp, Uint32 flags) {
    SDL_Surface* ret;

    if (!sdlhandle)
        sdlhack_init();
 
    if (force_res_x != -1) {
        width = force_res_x;
        height = force_res_y;
    }

    ret = orig_SDL_SetVideoMode(width, height, bpp, flags);
    if (ret)
        check_event_filter();

    if (!trigger_key)
        init_trigger();

    return ret;
}

void SDL_SetEventFilter(SDL_EventFilter filter) {
    app_event_filter = filter;
    check_event_filter();
}

SDL_EventFilter SDL_GetEventFilter() {
    return app_event_filter;
}

SDL_GrabMode SDL_WM_GrabInput(SDL_GrabMode gm) {
    if (gm != SDL_GRAB_QUERY) {
        app_grabmode = gm;
        if (!force_grab_off)
            orig_SDL_WM_GrabInput(app_grabmode);
    }
    return app_grabmode;
}

int SDL_NumJoysticks(void)
{

    if (orig_SDL_NumJoysticks)
        return orig_SDL_NumJoysticks();

    return 0;
}

SDL_Joystick *SDL_JoystickOpen(int device_index)
{
    if (orig_SDL_JoystickOpen)
        return orig_SDL_JoystickOpen(device_index);

    if (orig_SDL_SetError)
        orig_SDL_SetError("Joystick disabled");
    return NULL;
}

const char *SDL_JoystickName(int device_index)
{
    if (orig_SDL_JoystickName)
        return orig_SDL_JoystickName(device_index);
    if (orig_SDL_SetError)
        orig_SDL_SetError("Joystick disabled");
    return NULL;
}

int SDL_JoystickOpened(int device_index)
{
    if (orig_SDL_JoystickOpened)
        return orig_SDL_JoystickOpened(device_index);
    return 0;
}

/*
  Emulate the behavior of the original, but don't do anything.
 */

int SDL_JoystickEventState(int state)
{
    static int cstate = SDL_IGNORE;

    if (orig_SDL_JoystickEventState)
        return orig_SDL_JoystickEventState(state);

    if (state == SDL_QUERY)
        return cstate;
    cstate = state;
    return cstate;
}

